@@ -29,6 +29,11 @@ use aws_smithy_http::byte_stream::ByteStream;
 use futures::TryStreamExt;
 use http::HeaderValue;
 use http::StatusCode;
+use log::debug;
+use log::error;
+use log::info;
+use log::warn;
+use metrics::increment_counter;
 use once_cell::sync::Lazy;
 
 use super::error::parse_get_object_error;
@@ -63,11 +68,13 @@ static ENDPOINT_TEMPLATES: Lazy<HashMap<&'static str, &'static str>> = Lazy::new
     m
 });
 
+/// Builder for s3 services
+///
 /// # TODO
 ///
 /// enable_path_style and enable_signature_v2 need sdk support.
 ///
-/// ref: https://github.com/awslabs/aws-sdk-rust/issues/390
+/// ref: <https://github.com/awslabs/aws-sdk-rust/issues/390>
 #[derive(Default, Debug, Clone)]
 pub struct Builder {
     root: Option<String>,
@@ -75,8 +82,8 @@ pub struct Builder {
     bucket: String,
     credential: Option<Credential>,
     /// endpoint must be full uri, e.g.
-    /// - https://s3.amazonaws.com
-    /// - http://127.0.0.1:3000
+    /// - <https://s3.amazonaws.com>
+    /// - <http://127.0.0.1:3000>
     ///
     /// If user inputs endpoint like "s3.amazonaws.com", we will prepend
     /// "https://" before it.
@@ -117,6 +124,8 @@ impl Builder {
     }
 
     pub async fn finish(&mut self) -> Result<Arc<dyn Accessor>> {
+        info!("backend build started: {:?}", &self);
+
         let root = match &self.root {
             // Use "/" as root if user not specified.
             None => "/".to_string(),
@@ -131,6 +140,7 @@ impl Builder {
                 v
             }
         };
+        info!("backend use root {}", &root);
 
         // Handle endpoint, region and bucket name.
         let bucket = match self.bucket.is_empty() {
@@ -141,11 +151,13 @@ impl Builder {
                 source: anyhow!("bucket is empty"),
             }),
         }?;
+        debug!("backend use bucket {}", &bucket);
 
         let endpoint = match &self.endpoint {
             Some(endpoint) => endpoint,
             None => "https://s3.amazonaws.com",
         };
+        debug!("backend use endpoint {} to detect region", &endpoint);
 
         // Setup error context so that we don't need to construct many times.
         let mut context: HashMap<String, String> = HashMap::from([
@@ -223,6 +235,7 @@ impl Builder {
                 });
             }
         };
+        debug!("backend use endpoint: {}, region: {}", &endpoint, &region);
 
         // Config Loader will load config from environment.
         //
@@ -269,7 +282,9 @@ impl Builder {
                     ));
                 }
                 // We don't need to do anything if user tries to read credential from env.
-                Credential::Plain => {}
+                Credential::Plain => {
+                    warn!("backend got empty credential, fallback to read from env.")
+                }
                 _ => {
                     return Err(Error::Backend {
                         kind: Kind::BackendConfigurationInvalid,
@@ -288,8 +303,10 @@ impl Builder {
             .middleware(aws_smithy_client::erase::DynMiddleware::new(
                 DefaultMiddleware::new(),
             ))
+            .default_async_sleep()
             .build();
 
+        info!("backend build finished: {:?}", &self);
         Ok(Arc::new(Backend {
             root,
             bucket: self.bucket.clone(),
@@ -298,6 +315,7 @@ impl Builder {
     }
 }
 
+/// Backend for s3 services.
 #[derive(Debug, Clone)]
 pub struct Backend {
     bucket: String,
@@ -361,7 +379,13 @@ impl Backend {
 #[async_trait]
 impl Accessor for Backend {
     async fn read(&self, args: &OpRead) -> Result<BoxedAsyncReader> {
+        increment_counter!("opendal_s3_read_requests");
+
         let p = self.get_abs_path(&args.path);
+        info!(
+            "object {} read start: offset {:?}, size {:?}",
+            &p, args.offset, args.size
+        );
 
         let mut req = self
             .client
@@ -373,16 +397,22 @@ impl Accessor for Backend {
             req = req.range(HeaderRange::new(args.offset, args.size).to_string());
         }
 
-        let resp = req
-            .send()
-            .await
-            .map_err(|e| parse_get_object_error(e, "read", &p))?;
+        let resp = req.send().await.map_err(|e| {
+            let e = parse_get_object_error(e, "read", &p);
+            error!("object {} get_object: {:?}", &p, e);
+            e
+        })?;
 
+        info!(
+            "object {} reader created: offset {:?}, size {:?}",
+            &p, args.offset, args.size
+        );
         Ok(Box::new(S3ByteStream(resp.body).into_async_read()))
     }
 
     async fn write(&self, r: BoxedAsyncReader, args: &OpWrite) -> Result<usize> {
         let p = self.get_abs_path(&args.path);
+        info!("object {} write start: size {}", &p, args.size);
 
         let _ = self
             .client
@@ -395,13 +425,21 @@ impl Accessor for Backend {
             )))
             .send()
             .await
-            .map_err(|e| parse_unexpect_error(e, "write", &p))?;
+            .map_err(|e| {
+                let e = parse_unexpect_error(e, "write", &p);
+                error!("object {} put_object: {:?}", &p, e);
+                e
+            })?;
 
+        info!("object {} write finished: size {:?}", &p, args.size);
         Ok(args.size as usize)
     }
 
     async fn stat(&self, args: &OpStat) -> Result<Metadata> {
+        increment_counter!("opendal_s3_stat_requests");
+
         let p = self.get_abs_path(&args.path);
+        info!("object {} stat start", &p);
 
         let meta = self
             .client
@@ -426,6 +464,7 @@ impl Accessor for Backend {
 
                 m.set_complete();
 
+                info!("object {} stat finished", &p);
                 Ok(m)
             }
             // Always returns empty dir object if path is endswith "/" and we got an
@@ -437,14 +476,21 @@ impl Accessor for Backend {
                 m.set_mode(ObjectMode::DIR);
                 m.set_complete();
 
+                info!("object {} stat finished", &p);
                 Ok(m)
             }
-            Err(e) => Err(e),
+            Err(e) => {
+                error!("object {} head_object: {:?}", &p, e);
+                Err(e)
+            }
         }
     }
 
     async fn delete(&self, args: &OpDelete) -> Result<()> {
+        increment_counter!("opendal_s3_delete_requests");
+
         let p = self.get_abs_path(&args.path);
+        info!("object {} delete start", &p);
 
         let _ = self
             .client
@@ -455,15 +501,19 @@ impl Accessor for Backend {
             .await
             .map_err(|e| parse_unexpect_error(e, "delete", &p))?;
 
+        info!("object {} delete finished", &p);
         Ok(())
     }
 
     async fn list(&self, args: &OpList) -> Result<BoxedObjectStream> {
+        increment_counter!("opendal_s3_list_requests");
+
         let mut path = self.get_abs_path(&args.path);
         // Make sure list path is endswith '/'
         if !path.ends_with('/') && !path.is_empty() {
             path.push('/')
         }
+        info!("object {} list start", &path);
 
         Ok(Box::new(S3ObjectStream::new(
             self.clone(),
